/*
  This code is very much copied from bam2bcf.c and errmod.c from samtools 1.18

  All credit to auther of SAMtools!

 */

#include <cstdio>
#include <stdint.h>
#include <cstdlib>
#include <cmath>
#include "ctype.h"
#include "ksort.h"
#include "mUpPile.h"

KSORT_INIT_GENERIC(uint16_t)

#define ERR_DEP 0.83f

#define MIN_BASE_Q 13
#define CAPQ 60 //DRAGON SHOULD BE IMPLEMENTED
extern int refToInt[256];


typedef struct __errmod_coef_t {
	double *fk, *beta, *lhet;
} errmod_coef_t;



typedef struct {
	double depcorr;
	errmod_coef_t *coef;
} errmod_t;


typedef struct {
	double fsum[16], bsum[16];
	uint32_t c[16];
} call_aux_t;

int myIter =0;
static errmod_coef_t *cal_coef(double depcorr, double eta)
{
  //  fprintf(stderr,"\nmyIter=%d\t[%s] depcorr=%f eta=%f\n ",myIter,__FUNCTION__,depcorr,eta);
  int k, n, q;
  long double sum, sum1;
  double *lC;
  errmod_coef_t *ec;
  
  ec =(errmod_coef_t *) calloc(1, sizeof(errmod_coef_t));
  // initialize ->fk
  ec->fk = (double*)calloc(256, sizeof(double));
  ec->fk[0] = 1.0;
  for (n = 1; n != 256; ++n)
    ec->fk[n] = pow(1. - depcorr, n) * (1.0 - eta) + eta;
  // initialize ->coef
  ec->beta = (double*)calloc(256 * 256 * 64, sizeof(double));
  lC = (double*)calloc(256 * 256, sizeof(double));
  for (n = 1; n != 256; ++n) {
    double lgn = lgamma(n+1);
    for (k = 1; k <= n; ++k)
      lC[n<<8|k] = lgn - lgamma(k+1) - lgamma(n-k+1);
  }
  for (q = 1; q != 64; ++q) {
    double e = pow(10.0, -q/10.0);
    double le = log(e);
    double le1 = log(1.0 - e);
    for (n = 1; n <= 255; ++n) {
      double *beta = ec->beta + (q<<16|n<<8);
      sum1 = sum = 0.0;
      for (k = n; k >= 0; --k, sum1 = sum) {
	sum = sum1 + expl(lC[n<<8|k] + k*le + (n-k)*le1);
	beta[k] = -10. / M_LN10 * logl(sum1 / sum);
      }
    }
  }
  // initialize ->lhet
  ec->lhet = (double*)calloc(256 * 256, sizeof(double));
  for (n = 0; n < 256; ++n)
    for (k = 0; k < 256; ++k)
      ec->lhet[n<<8|k] = lC[n<<8|k] - M_LN2 * n;
  free(lC);
  return ec;
}

errmod_t *errmod_init(float depcorr) {
  //  fprintf(stderr,"[%s]",__FUNCTION__);
  errmod_t *em;
  em = (errmod_t*)calloc(1, sizeof(errmod_t));
  em->depcorr = depcorr;
  em->coef = cal_coef(depcorr, 0.03);
  return em;
}

void errmod_destroy(errmod_t *em)
{
	if (em == 0) return;
	free(em->coef->lhet); free(em->coef->fk); free(em->coef->beta);
	free(em->coef); free(em);
}



// qual:6, strand:1, base:4
int errmod_cal(const errmod_t *em, int n, int m, uint16_t *bases, float *q)
{
  myIter++;
  //  fprintf(stderr,"\nmyIter=%d\t[%s] n=%d m=%d\n",myIter,__FUNCTION__,n,m );
  
  call_aux_t aux;
  int i, j, k, w[32];
  
  for(i=0;0&&i<n;i++)
    fprintf(stderr,"\t->[%s][%d] %d\n",__FUNCTION__,i,bases[i]);
  //  fprintf(stderr,"\n");
  if (m > m) return -1;
  memset(q, 0, m * m * sizeof(float));
  if (n == 0) return 0;
  // calculate aux.esum and aux.fsum
  if (n > 255) { // then sample 255 bases
    ks_shuffle(uint16_t, n, bases);
    n = 255;
  }
  ks_introsort(uint16_t, n, bases);
  memset(w, 0, 32 * sizeof(int));
  memset(&aux, 0, sizeof(call_aux_t));
  for (j = n - 1; j >= 0; --j) { // calculate esum and fsum
    uint16_t b = bases[j];
    int q = b>>5 < 4? 4 : b>>5;
    if (q > 63) q = 63;
    k = b&0x1f;
    aux.fsum[k&0xf] += em->coef->fk[w[k]];
    aux.bsum[k&0xf] += em->coef->fk[w[k]] * em->coef->beta[q<<16|n<<8|aux.c[k&0xf]];
    ++aux.c[k&0xf];
    ++w[k];
  }
  // generate likelihood
  for (j = 0; j != m; ++j) {
    float tmp1, tmp3;
    int tmp2, bar_e;
    // homozygous
    for (k = 0, tmp1 = tmp3 = 0.0, tmp2 = 0; k != m; ++k) {
      if (k == j) continue;
      tmp1 += aux.bsum[k]; tmp2 += aux.c[k]; tmp3 += aux.fsum[k];
    }
    if (tmp2) {
      bar_e = (int)(tmp1 / tmp3 + 0.499);
      if (bar_e > 63) bar_e = 63;
      q[j*m+j] = tmp1;
    }
    // heterozygous
    for (k = j + 1; k < m; ++k) {
      int cjk = aux.c[j] + aux.c[k];
      for (i = 0, tmp2 = 0, tmp1 = tmp3 = 0.0; i < m; ++i) {
	if (i == j || i == k) continue;
	tmp1 += aux.bsum[i]; tmp2 += aux.c[i]; tmp3 += aux.fsum[i];
      }
      if (tmp2) {
	bar_e = (int)(tmp1 / tmp3 + 0.499);
	if (bar_e > 63) bar_e = 63;
	q[j*m+k] = q[k*m+j] = -4.343 * em->coef->lhet[cjk<<8|aux.c[k]] + tmp1;
      } else q[j*m+k] = q[k*m+j] = -4.343 * em->coef->lhet[cjk<<8|aux.c[k]]; // all the bases are either j or k
    }
    for (k = 0; k != m; ++k) if (q[j*m+k] < 0.0) q[j*m+k] = 0.0;
  }
  //  exit(0);
  return 0;
}

void call_bam(chunkyT *chk,double **lk){
  static errmod_t *mod = NULL;
  if(chk==NULL){
    errmod_destroy(mod);    
    return;
  }
  if(mod==NULL)
    mod = errmod_init(1-ERR_DEP);
  
  for(int s=0;s<chk->nSites;s++){
    //    fprintf(stderr,"%d:\n",s);
    lk[s] = new double[10*chk->nSamples];
    for(int ii=0;ii<10*chk->nSamples;ii++)
      lk[s][ii] = -0.0;
    //fprintf(stderr,"chr=%s minbaseq=%d\n",chk->refName,MIN_BASE_Q);
    for(int i=0;i<chk->nSamples;i++){

      tNode &nd = chk->nd[s][i];
      uint16_t bases[nd.l];
      int numItems =0;
      for(int j=0;j<nd.l;j++) {
	int q=nd.qs[j];
	//	fprintf(stderr,"q=%d\n",q);
	if(q<MIN_BASE_Q){
	  //	  fprintf(stderr,"skipping due to low stuff\n");
	  continue;
	}

	
	if (q>nd.mapQ[j])
	  q=nd.mapQ[j];
	if (q > 63) q = 63;
	if (q < 4) q = 4;
	//	fprintf(stderr,"q=%d qshift=%d\n",q,q<<5);
	//plug in values
	//	bca->bases[n++] = q<<5 | (int)bam1_strand(p->b)<<4 | b;
	int isLower = islower(nd.seq[j])>0;
	//	fprintf(stderr,"strandstuff=%d\tshiftstrandstuff=%d\n",isLower, isLower<<4 );
	bases[numItems++] = q<<5 | isLower<<4 | refToInt[nd.seq[j]];
	//	fprintf(stderr,"bases[%d]=%d\n",j,bases[numItems-1]);
      }
      float likes[25];
      for(int j=0;0&&j<25;j++) 
	likes[j]=0;
      if(numItems!=0) {
	errmod_cal(mod,numItems,5,bases,likes);
	int ord[] = {0,1,2,3,6,7,8,12,13,18};
	for(int gg=0;gg<10;gg++){
	  //	  fprintf(stderr,"%f\n",likes[ord[gg]]);
	  lk[s][i*10+gg] = -log(10.0) *likes[ord[gg]]/10.0;
	}
	/*
	fprintf(fpp,"lik:\t");
	for(int n=0;n<5;n++){
	  for(int m=0;m<5;m++)
	    fprintf(fpp,"%f ",likes[n*5+m]);
	  
	} 
	fprintf(fpp,"\n");

	fflush(fpp);
	//	exit(0);
	*/
      }
    }
    //    break;
  }
}
