/*
  daemon/dragon uniqueonly is externed from filereading
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cassert>
#include <ctype.h>
#include <map>
#include "bams.h"
#include "bgzf.h"
#include "printer.h"

extern int uniqueOnly;

uint32_t bam_calend(const aRead& rd, const uint32_t *cigar){
  uint32_t end = rd.pos;
  //  fprintf(stderr,"end start=%d\n",end);
  for (int k = 0; k < rd.nCig; ++k) {
    int op = cigar[k] & BAM_CIGAR_MASK;
    if (op == BAM_CMATCH || op == BAM_CDEL || op == BAM_CREF_SKIP)
      end += cigar[k] >> BAM_CIGAR_SHIFT;
  }
  //  fprintf(stderr,"end is : %d\n",end);
  return end;
}


int is_overlap(uint32_t beg, uint32_t end, const aRead &b) {
	uint32_t rbeg = b.pos;
	uint32_t rend = b.nCig ? bam_calend(b, getCig(&b)) : b.pos + 1;
	int ret = (rend > beg && rbeg < end);
	return ret;
}



int bam_validate1(const aHead *header, const aRead b)
{
  
  if (b.refID < -1 || b.next_refID < -1){
    fprintf(stderr,"error first\n");
    return 0;
  }
  if (header && (b.refID >= header->n_ref || b.next_refID >= header->n_ref)){
    fprintf(stderr,"error second\n");
    return 0;
  }
  //  int l_qname = b.bin_mq_nl &0xff;

  if (b.block_size < b.l_qname){
      fprintf(stderr,"error third\n");
      return 0;

    }
    char *s =(char *) memchr(b.vDat, '\0', b.l_qname);
 
    //    fprintf(stderr,"%d\n",(int )s-(b.vDat+1));
    if (s[0] != b.vDat[b.l_qname-1]) {
      fprintf(stderr,"error forth\n");
    return 0;
    }
  // FIXME: Other fields could also be checked, especially the auxiliary data
      
      return 1;
}


 void printIter(const iter_t& it,FILE *fp){
   fprintf(stderr,"--------------\n");
   fprintf(stderr,"from_first=%d\n",it.from_first);
   fprintf(stderr,"tid=%d\tbeg=%d\tend=%d\tn_off=%d\ti=%d\tfinished=%d\n",it.tid,it.beg,it.end,it.n_off,it.i,it.finished);
   fprintf(stderr,"i=%d\tcurr_off:%lu\n",it.i,it.curr_off);
   for(int i=0;i<it.n_off;i++)
     fprintf(stderr,"%d: (%lu-%lu)\n",i,it.off[i].chunk_beg,it.off[i].chunk_end);
   fprintf(stderr,"--------------\n");
 }




aHead *getHd(BGZF *gz){
  aHead *h = new aHead;
  
  const char * magic ="BAM\1"; 
  char head[4];
  if(4!=bgzf_read(gz,head,4)){
    fprintf(stderr,"[%s] Error reading\n",__FUNCTION__);
  }
  //  print(stderr,head,4);
  if(strncmp(head,magic,4)!=0) {
    fprintf(stderr,"[%s]\t-> problem with magic number\n",__FUNCTION__);
    exit(0);
  }
  bgzf_read(gz,&h->l_text,sizeof(int));
  h->text = new char[h->l_text];
  bgzf_read(gz,h->text,h->l_text);
  bgzf_read(gz,&h->n_ref,sizeof(int));
  h->l_name=new int[h->n_ref];
  h->name = new char*[h->n_ref];
  h->l_ref = new int[h->n_ref];
  
  for(int i=0;i<h->n_ref;i++){
    bgzf_read(gz,&h->l_name[i],sizeof(int));
    h->name[i] = new char[h->l_name[i]];
    bgzf_read(gz,h->name[i],h->l_name[i]);
    bgzf_read(gz,&h->l_ref[i],sizeof(int));
  } 
  return h;
}



/*
  given the current offset read an alignment and put the data in st
 */
int getAlign(BGZF *gz,int block_size,aRead &st) {

  uint32_t tmp[8];
  bgzf_read(gz,tmp,32);
  //copy vals to st
  st.refID = tmp[0];
  st.pos = tmp[1];
  st.bin_mq_nl = tmp[2];
  st.flag_nc = tmp[3];
  st.l_seq = tmp[4];
  st.next_refID = tmp[5];
  st.next_pos = tmp[6];
  st.tlen = tmp[7];
  
  if(0){//debug print
    fprintf(stderr,"refID:%d\n",st.refID);
    fprintf(stderr,"pos:%d\n",st.pos+1);
    int nBin = st.bin_mq_nl >>16;
    int mapQ = st.bin_mq_nl >>8&0xff;
    int l_qname = st.bin_mq_nl &0xff;
    int flag = st.flag_nc >>16;
    int nCig = st.flag_nc &0xffff;
    
    fprintf(stderr,"nbin=%u mapQ=%u l_qname=%u flag=%u nCig=%u\n",nBin,mapQ,l_qname,flag,nCig);
    fprintf(stderr,"l_seq:%d\n",st.l_seq);
    fprintf(stderr,"nnext_refID:%d\n",st.next_refID);
    fprintf(stderr,"next_pos:%d\n",st.next_pos);
    fprintf(stderr,"tlen:%d\n",st.tlen);
  }
  //begin vars
  // fprintf(stderr,"reading\t%d\n",block_size-32);
  if(block_size-32>=RLEN){//fix if blocks to big
    delete [] st.vDat;
    int tmpSize = block_size-32;
    kroundup32(tmpSize);
    st.vDat = new uint8_t[tmpSize];
  }
    
  bgzf_read(gz,st.vDat,block_size-32);
  st.block_size=block_size-32;
  //  fprintf(stderr,"st.vDat=%d\n",block_size-32);
  
  //plugin bitparsed attributes maybe we don't want to do this...
  st.nBin = st.bin_mq_nl >>16;
  st.mapQ = st.bin_mq_nl >>8&0xff;
  st.l_qname = st.bin_mq_nl &0xff;
  st.flag = st.flag_nc >>16;
  st.nCig = st.flag_nc &0xffff;
  return 0;
}


int bam_read1(BGZF *fp,aRead & b){

  int block_size;
  if(4!=bgzf_read(fp,&block_size,sizeof(int))){
    fprintf(stderr,"[%s] error reading some file stuff in assuming EOF\n",__FUNCTION__);
    //    return 0;
    return -1;
  }
  getAlign(fp,block_size,b);
  return 1;
}



int getNumBest(uint8_t *s, uint8_t *sStop) {
  //  fprintf(stderr,"\ncomp:%p vs %p\n",s,sStop);
  int retVal =1;
  while (s < sStop) {
    uint8_t type;
    int doRet =0;
    if(s[0]=='X'&&s[1]=='0')
      doRet = 1;
	//    fprintf(stderr,"\t%c%c:",s[0],s[1]);
    s += 2; type = *s; ++s;
    //    fprintf(stderr,"\ntype=%c\n",type);//,(char)*s);
    //    kputc('\t', &str); kputsn((char*)key, 2, &str); kputc(':', &str);
    if (type == 'A') {if(doRet) return (int) *s ;else  ++s; }
    else if (type == 'C') {if(doRet) return *s;else ++s; }
    else if (type == 'c') {if(doRet) return *(int8_t*)s; else  ++s; }
    else if (type == 'S') {if(doRet) return *(uint16_t*)s; else s += 2; }
    else if (type == 's') {if(doRet) return *(int16_t*)s; else  s += 2; }
    else if (type == 'I') {if(doRet) return *(uint32_t*)s;else  s += 4; }
    else if (type == 'i') {if(doRet) return *(int32_t*)s; else  s += 4; }
    else if (type == 'f') { s += 4; continue; }
    else if (type == 'd') {s += 8;continue; }
    else if (type == 'Z' || type == 'H') return retVal;
  }
  //  fprintf(stderr,"done\n");
}


int bam_read2(BGZF *fp,aRead & b){
 reread: //oh yes baby a goto statement
  int block_size;
  if(4!=bgzf_read(fp,&block_size,sizeof(int))){
    //    fprintf(stderr,"[%s] error reading some file stuff in assuming EOF\n",__FUNCTION__);
    //    return 0;
    return -2;
  }
  getAlign(fp,block_size,b);
  //  fprintf(stderr,"b.pos=%d\t",b.pos);
  //  return 1;//SOAP DRAGON
  if(uniqueOnly==0)
    return 1;
  int nBest = getNumBest(getAuxStart(&b),b.vDat+b.block_size);
  if(0&&!(b.flag&BAM_FPROPER_PAIR)){
    fprintf(stderr,"[%s] unproper read will skip:\n",__FUNCTION__);
    goto reread;
  }
  //validate the number of optimal hits
  //  fprintf(stderr,"pos=%d\t",b.pos);

  if(nBest!=1){
    //fprintf(stderr,"nBest=%d\tpos=%d\n",nBest,b.pos);
    goto reread;
  }
  //  fprintf(stderr,"\n");
  return 1;
}


void dalloc(aHead *hd){
  delete [] hd->text;
  delete [] hd->l_name;
  delete [] hd->l_ref;
  for(int i=0;i<hd->n_ref;i++)
    delete [] hd->name[i];
  delete [] hd->name;
  delete hd;
  hd=NULL;
}


BGZF *openBAM(const char *fname){
  BGZF* fp =NULL;
  if((fp=bgzf_open(fname,"r"))==NULL ){
    fprintf(stderr,"nonexistant file: %s\n",fname);
    exit(0);
  }
  const char *str = strrchr(fname,'.');
  if(str&&strcmp(str,".bam")!=0){
    fprintf(stderr,"file:\"%s\" should be suffixed with \".bam\"\n",fname);
    exit(0);
  }
  return fp;
}



int bam_iter_read(BGZF *fp, iter_t *iter, aRead &b) {
  //  fprintf(stderr,"[%s] it from_first=%d iter->finished=%d\n",__FUNCTION__,iter->from_first,iter->finished);
  int ret;
  if (iter && iter->finished) return -1;
  if (iter == 0 || iter->from_first) {
    //fprintf(stderr,"[%s] reading\n",__FUNCTION__);
    ret = bam_read1(fp, b);
    if (ret < 0 && iter) iter->finished = 1;
    return ret;
  }
  if (iter->off == 0) return -1;
  for (;;) {
    if (iter->curr_off == 0 || iter->curr_off >= iter->off[iter->i].chunk_end) { // then jump to the next chunk
      if (iter->i == iter->n_off - 1) {
	ret = -1; break; 
      } // no more chunks
      if (iter->i >= 0) assert(iter->curr_off == iter->off[iter->i].chunk_end); // otherwise bug
      if (iter->i < 0 || iter->off[iter->i].chunk_end != iter->off[iter->i+1].chunk_beg) { // not adjacent chunks; then seek
	bgzf_seek(fp, iter->off[iter->i+1].chunk_beg, SEEK_SET);
	iter->curr_off = bgzf_tell(fp);
      }
      ++iter->i;
    }
    if ((ret = bam_read1(fp, b)) >= 0) {
      iter->curr_off = bgzf_tell(fp);
      if (b.refID != iter->tid || b.pos >= iter->end) { // no need to proceed
	ret = bam_validate1(NULL, b)? -1 : -5; // determine whether end of region or error
	break;
      }
      else if (is_overlap(iter->beg, iter->end, b)) return ret;
    } else break; // end of file or error
  }
  iter->finished = 1;
  return ret;
}


int bam_iter_read1(BGZF *fp, iter_t *iter, aRead &b) {
  //  fprintf(stderr,"[%s] it from_first=%d iter->finished=%d\n",__FUNCTION__,iter->from_first,iter->finished);
  int ret;
  if (iter && iter->finished) return -2;
  if (iter == 0 || iter->from_first) {
    //fprintf(stderr,"[%s] reading\n",__FUNCTION__);
    ret = bam_read2(fp, b);
    if (ret < 0 && iter) iter->finished = 1;
    return ret;
  }
  if (iter->off == 0) return -1;
  for (;;) {
    if (iter->curr_off == 0 || iter->curr_off >= iter->off[iter->i].chunk_end) { // then jump to the next chunk
      if (iter->i == iter->n_off - 1) {
	ret = -1; break; 
      } // no more chunks
      if (iter->i >= 0) assert(iter->curr_off == iter->off[iter->i].chunk_end); // otherwise bug
      if (iter->i < 0 || iter->off[iter->i].chunk_end != iter->off[iter->i+1].chunk_beg) { // not adjacent chunks; then seek
	bgzf_seek(fp, iter->off[iter->i+1].chunk_beg, SEEK_SET);
	iter->curr_off = bgzf_tell(fp);
      }
      ++(iter->i);
    }
    //    fprintf(stderr,"in the for looop2 curr_off=%lu\n",iter->curr_off);
    if ((ret = bam_read2(fp, b)) >= 0) {
      iter->curr_off = bgzf_tell(fp);
      //      fprintf(stderr,"in the for looop3 curr_off=%lu\n",iter->curr_off);
      if (b.refID != iter->tid || b.pos >= iter->end) { // no need to proceed
	//	fprintf(stderr,"in the for looop4 no ned to procedd:%lu\n",iter->curr_off);
	//	fprintf(stderr,"bamret=%d\n", bam_validate1(NULL, b));
	ret = bam_validate1(NULL, b)? -1 : -5; // determine whether end of region or error

	break;
      }
      else if (is_overlap(iter->beg, iter->end, b)) return ret;
    } else break; // end of file or error
  }
  //  fprintf(stderr,"asdf;laksdf;alsdkfj ret=%d\n",ret);
  iter->finished = 1;
  return ret;
}


/*

void printNode(const node &nd,FILE *fp,int printDepth,const char *chr){
  //  fprintf(stderr,"printing node with len:%d\n",nd.len);
  fprintf(fp,"%s\t",chr);
  fprintf(fp,"%d\t",nd.refPos+1);

  fprintf(fp,"%d\t",nd.depth);
  for(size_t i=0;i<nd.seq.l;i++)
    fprintf(fp,"%c",nd.seq.s[i]);
  fprintf(fp,"\t");
  
  for(size_t i=0;i<nd.qs.l;i++)
    fprintf(fp,"%c",nd.qs.s[i]);

  if(printDepth ){
    fprintf(fp,"\t");
    for(size_t i=0;1&&i<nd.pos.l;i++)
      fprintf(fp,"%c",nd.pos.s[i]);
  }
  fprintf(fp,"\n");
  free(nd.seq.s);
  free(nd.qs.s);
  free(nd.pos.s);
}

*/

void dalloc (sglPool &ret){
  for(int i=0;i<ret.l;i++)
    delete [] ret.reads[i].vDat;
  if(0){
    free( ret.reads);
    free(ret.first);
    free(ret.last);
  }else{
    delete [] ret.reads;
    delete [] ret.first;
    delete [] ret.last;


  }
}
